@@ -1,9 +1,9 @@
 import requests
 import asyncio
-import websockets
 import uuid
 import json
 import datetime
+import aiohttp
 
 import logging
 
@@ -14,6 +14,7 @@ TRANS_SERVER = "wss://v1-linkplus-app.lightwaverf.com"
 VERSION = "1.6.8"
 MAX_RETRIES = 5
 PUBLIC_AUTH_SERVER = "https://auth.lightwaverf.com/token"
+PUBLIC_API = "https://publicapi.lightwaverf.com/v1/"
 
 class _LWRFMessage:
     _tran_id = 0
@@ -43,7 +44,6 @@ class _LWRFMessageItem:
         _LWRFMessageItem._item_id += 1
         self._item["payload"] = payload
 
-
 class _LWRFFeatureSet:
 
     def __init__(self):
@@ -68,17 +68,23 @@ class _LWRFFeatureSet:
     def is_gen2(self):
         return self._gen2
 
-
 class LWLink2:
 
-    def __init__(self, username, password):
+    def __init__(self, username=None, password=None):
+
+        self.featuresets = {}
+        self._authtoken = None
+
         self._username = username
         self._password = password
+
+        self._session = aiohttp.ClientSession()
+
+        #Websocket only variables:
         self._device_id = str(uuid.uuid4())
-        self._authtoken = None
         self._websocket = None
         self._callback = []
-        self.featuresets = []
+        self._group_ids = []
 
         # Next three variables are used to synchronise responses to requests
         self._transaction = None
@@ -87,19 +93,15 @@ class LWLink2:
 
         asyncio.ensure_future(self._consumer_handler())
 
-    def _sendmessage(self, message):
-        return asyncio.get_event_loop().run_until_complete(self._async_sendmessage(message))
-
     async def _async_sendmessage(self, message, _retry=1):
-        # await self.outgoing.put(message)
 
-        if not self._websocket:
+        if not self._websocket or self._websocket.closed:
             _LOGGER.debug("Can't send (websocket closed), reconnecting")
             await self.async_connect()
             _LOGGER.debug("Connection reopened")
 
         _LOGGER.debug("Sending: %s", message.json())
-        await self._websocket.send(message.json())
+        await self._websocket.send_str(message.json())
         self._transaction = message._message["transactionId"]
         self._waitingforresponse.clear()
         await self._waitingforresponse.wait()
@@ -121,52 +123,60 @@ class LWLink2:
     def _consumer_handler(self):
         while True:
             try:
-                jsonmessage = yield from self._websocket.recv()
-                message = json.loads(jsonmessage)
-                _LOGGER.debug("Received %s", message)
-                # Some transaction IDs don't work, this is a workaround
-                if message["class"] == "feature" and (
-                        message["operation"] == "write" or message["operation"] == "read"):
-                    message["transactionId"] = message["items"][0]["itemId"]
-                # now parse the message
-                if message["transactionId"] == self._transaction:
+                mess = yield from self._websocket.receive()
+                _LOGGER.debug("Received %s", mess)
+                if mess.type == aiohttp.WSMsgType.TEXT:
+                    message = mess.json()
+                    # Some transaction IDs don't work, this is a workaround
+                    if message["class"] == "feature" and (
+                            message["operation"] == "write" or message["operation"] == "read"):
+                        message["transactionId"] = message["items"][0]["itemId"]
+                    # now parse the message
+                    if message["transactionId"] == self._transaction:
+                        self._waitingforresponse.set()
+                        self._transactions = None
+                        self._response = message
+                    elif message["direction"] == "notification" and message["class"] == "group" \
+                            and message["operation"] == "event":
+                        yield from self.async_get_hierarchy()
+                    elif message["direction"] == "notification" and message["operation"] == "event":
+                        #TODO: First call is unecessary, see if second works then refactor the first one out
+                        if "_feature" in message["items"][0]["payload"]:
+                            feature_id = message["items"][0]["payload"]["_feature"]["featureId"]
+                            feature = message["items"][0]["payload"]["_feature"]["featureType"]
+                            value = message["items"][0]["payload"]["value"]
+                            self.get_featureset_by_featureid(feature_id).features[feature][1] = value
+                            _LOGGER.debug("Event with _feature received (%s %s %s), calling callbacks %s", feature_id, feature, value, self._callback)
+                            for func in self._callback:
+                                func()
+                        elif "featureId" in message["items"][0]["payload"]:
+                            feature_id = message["items"][0]["payload"]["featureId"]
+                            feature = self.get_feature_by_featureid(feature_id)
+                            value = message["items"][0]["payload"]["value"]
+                            self.get_featureset_by_featureid(feature_id).features[feature][1] = value
+                            _LOGGER.debug("Event with _feature received (%s %s %s), calling callbacks %s", feature_id, feature, value, self._callback)
+                            for func in self._callback:
+                                func()
+                        else:
+                            _LOGGER.warning("Unhandled event message: %s", message)
+                    else:
+                        _LOGGER.warning("Received unhandled message: %s", message)
+                elif mess.type == aiohttp.WSMsgType.CLOSED:
+                # We're not going to get a response, so clear response flag to allow _send_message to unblock
+                    _LOGGER.debug("Websocket closed in message handler")
                     self._waitingforresponse.set()
                     self._transactions = None
-                    self._response = message
-                elif message["direction"] == "notification" and message["class"] == "group" \
-                        and message["operation"] == "event":
-                    yield from self.async_get_hierarchy()
-                elif message["direction"] == "notification" and message["operation"] == "event":
-                    if "_feature" in message["items"][0]["payload"]:
-                        feature_id = message["items"][0]["payload"]["_feature"]["featureId"]
-                        feature = message["items"][0]["payload"]["_feature"]["featureType"]
-                        value = message["items"][0]["payload"]["value"]
-                        self.get_featureset_by_featureid(feature_id).features[feature][1] = value
-                        _LOGGER.debug("Calling callbacks %s", self._callback)
-                        for func in self._callback:
-                            func()
-                    else:
-                        _LOGGER.warning("Message with no _feature: %s", message)
-                else:
-                    _LOGGER.warning("Received unhandled message: %s", message)
+                    self._response = None
+                    self._websocket = None
+                    asyncio.ensure_future(self.async_connect())
             except AttributeError:  # websocket is None if not set up, just wait for a while
                 yield from asyncio.sleep(1)
-            except websockets.ConnectionClosed:
-                # We're not going to get a response, so clear response flag to allow _send_message to unblock
-                _LOGGER.debug("Websocket closed in message handler")
-                self._waitingforresponse.set()
-                self._transactions = None
-                self._response = None
-                self._websocket = None
-                yield from asyncio.sleep(1)
+
 
     async def async_register_callback(self, callback):
         _LOGGER.debug("Register callback %s", callback)
         self._callback.append(callback)
 
-    def get_hierarchy(self):
-        return asyncio.get_event_loop().run_until_complete(self.async_get_hierarchy())
-
     async def async_get_hierarchy(self):
         _LOGGER.debug("Reading hierarchy")
         readmess = _LWRFMessage("user", "rootGroups")
@@ -174,15 +184,18 @@ class LWLink2:
         readmess.additem(readitem)
         response = await self._async_sendmessage(readmess)
 
+        self._group_ids = []
         for item in response["items"]:
-            group_ids = item["payload"]["groupIds"]
-            _LOGGER.debug("Reading groups {}".format(group_ids))
-            await self._async_read_groups(group_ids)
+            self._group_ids = self._group_ids + item["payload"]["groupIds"]
+
+        _LOGGER.debug("Reading groups {}".format(self._group_ids))
+        await self._async_read_groups()
 
         await self.async_update_featureset_states()
 
-    async def _async_read_groups(self, group_ids):
-        for groupId in group_ids:
+    async def _async_read_groups(self):
+        self.featuresets = {}
+        for groupId in self._group_ids:
             readmess = _LWRFMessage("group", "read")
             readitem = _LWRFMessageItem({"groupId": groupId,
                                          "blocks": True,
@@ -194,31 +207,19 @@ class LWLink2:
             readmess.additem(readitem)
             response = await self._async_sendmessage(readmess)
 
-            self.featuresets = []
-            for x in list(response["items"][0]["payload"]["devices"].values()):
-                for y in x["featureSetGroupIds"]:
-                    _LOGGER.debug("Creating device {}".format(x))
-                    new_featureset = _LWRFFeatureSet()
-                    new_featureset.featureset_id = y
-                    new_featureset.product_code = x["productCode"]
-
-                    readmess = _LWRFMessage("group", "read")
-                    readitem = _LWRFMessageItem({"groupId": y,
-                                                 "blocks": True,
-                                                 "devices": True,
-                                                 "features": True,
-                                                 "scripts": True,
-                                                 "subgroups": True,
-                                                 "subgroupDepth": 10})
-                    readmess.additem(readitem)
-                    featgroupresponse = await self._async_sendmessage(readmess)
-                    new_featureset.name = featgroupresponse["items"][0]["payload"]["name"]
-                    self.featuresets.append(new_featureset)
-
             for x in list(response["items"][0]["payload"]["features"].values()):
                 for z in x["groups"]:
+
+                    if z not in self.featuresets:
+                        _LOGGER.debug("Creating device {}".format(x))
+                        new_featureset = _LWRFFeatureSet()
+                        new_featureset.featureset_id = z
+                        new_featureset.product_code = "Not working" #TODO!
+                        new_featureset.name = x["name"]
+                        self.featuresets[z] = new_featureset
+
                     _LOGGER.debug("Adding device features {}".format(x))
-                    y = self.get_featureset_by_id(z)
+                    y = self.featuresets[z]
                     y.features[x["attributes"]["type"]] = [x["featureId"], x["attributes"]["value"]]
                     if x["attributes"]["type"] == "switch":
                         y._switchable = True
@@ -229,27 +230,18 @@ class LWLink2:
                     if x["attributes"]["type"] == "identify":
                         y._gen2 = True
 
-    def update_featureset_states(self):
-        return asyncio.get_event_loop().run_until_complete(self.async_update_featureset_states())
-
     async def async_update_featureset_states(self):
-        for x in self.featuresets:
+        for dummy, x in self.featuresets.items():
             for y in x.features:
                 value = await self.async_read_feature(x.features[y][0])
                 x.features[y][1] = value["items"][0]["payload"]["value"]
 
-    def write_feature(self, feature_id, value):
-        return asyncio.get_event_loop().run_until_complete(self.async_write_feature(feature_id, value))
-
     async def async_write_feature(self, feature_id, value):
         readmess = _LWRFMessage("feature", "write")
         readitem = _LWRFMessageItem({"featureId": feature_id, "value": value})
         readmess.additem(readitem)
         await self._async_sendmessage(readmess)
 
-    def read_feature(self, feature_id):
-        return asyncio.get_event_loop().run_until_complete(self.async_read_feature(feature_id))
-
     async def async_read_feature(self, feature_id):
         readmess = _LWRFMessage("feature", "read")
         readitem = _LWRFMessageItem({"featureId": feature_id})
@@ -257,45 +249,36 @@ class LWLink2:
         return await self._async_sendmessage(readmess)
 
     def get_featureset_by_id(self, featureset_id):
-        for x in self.featuresets:
-            if x.featureset_id == featureset_id:
-                return x
-        return None
+        return self.featuresets[featureset_id]
 
     def get_featureset_by_featureid(self, feature_id):
-        for x in self.featuresets:
+        for dummy, x in self.featuresets.items():
             for y in x.features.values():
                 if y[0] == feature_id:
                     return x
         return None
 
-    def turn_on_by_featureset_id(self, featureset_id):
-        return asyncio.get_event_loop().run_until_complete(self.async_turn_on_by_featureset_id(featureset_id))
+    def get_feature_by_featureid(self, feature_id):
+        for dummy, x in self.featuresets.items():
+            for z, y in x.features.items():
+                if y[0] == feature_id:
+                    return z
 
     async def async_turn_on_by_featureset_id(self, featureset_id):
         y = self.get_featureset_by_id(featureset_id)
         feature_id = y.features["switch"][0]
         await self.async_write_feature(feature_id, 1)
 
-    def turn_off_by_featureset_id(self, featureset_id):
-        return asyncio.get_event_loop().run_until_complete(self.async_turn_off_by_featureset_id(featureset_id))
-
     async def async_turn_off_by_featureset_id(self, featureset_id):
         y = self.get_featureset_by_id(featureset_id)
         feature_id = y.features["switch"][0]
         await self.async_write_feature(feature_id, 0)
 
-    def set_brightness_by_featureset_id(self, featureset_id, level):
-        return asyncio.get_event_loop().run_until_complete(self.async_set_brightness_by_featureset_id(featureset_id, level))
-
     async def async_set_brightness_by_featureset_id(self, featureset_id, level):
         y = self.get_featureset_by_id(featureset_id)
         feature_id = y.features["dimLevel"][0]
         await self.async_write_feature(feature_id, level)
 
-    def set_temperature_by_featureset_id(self, featureset_id, level):
-        return asyncio.get_event_loop().run_until_complete(self.async_set_temperature_by_featureset_id(featureset_id, level))
-
     async def async_set_temperature_by_featureset_id(self, featureset_id, level):
         y = self.get_featureset_by_id(featureset_id)
         feature_id = y.features["targetTemperature"][0]
@@ -303,21 +286,21 @@ class LWLink2:
 
     def get_switches(self):
         temp = []
-        for x in self.featuresets:
+        for dummy, x in self.featuresets.items():
             if x.is_switch():
                 temp.append((x.featureset_id, x.name))
         return temp
 
     def get_lights(self):
         temp = []
-        for x in self.featuresets:
+        for dummy, x in self.featuresets.items():
             if x.is_light():
                 temp.append((x.featureset_id, x.name))
         return temp
 
     def get_climates(self):
         temp = []
-        for x in self.featuresets:
+        for dummy, x in self.featuresets.items():
             if x.is_climate():
                 temp.append((x.featureset_id, x.name))
         return temp
@@ -326,13 +309,10 @@ class LWLink2:
     # Connection
     #########################################################
 
-    def connect(self):
-        return asyncio.get_event_loop().run_until_complete(self.async_connect())
-
     async def async_connect(self, tries=0):
         try:
-            if not self._websocket or not self._websocket.open:
-                self._websocket = await websockets.connect(TRANS_SERVER, ssl=True)
+            if not self._websocket or self._websocket.closed:
+                self._websocket = await self._session.ws_connect(TRANS_SERVER)
             return await self._authenticate()
         except Exception as exp:
             retry_delay = min(2 ** (tries + 1), 120)
@@ -342,7 +322,7 @@ class LWLink2:
 
     async def _authenticate(self):
         if not self._authtoken:
-            self._authtoken = self._get_access_token()
+            await self._get_access_token()
         if self._authtoken:
             authmess = _LWRFMessage("user", "authenticate")
             authpayload = _LWRFMessageItem({"token": self._authtoken, "clientDeviceId": self._device_id})
@@ -364,41 +344,89 @@ class LWLink2:
         else:
             return None
 
-    def _get_access_token(self):
-        _LOGGER.debug("Requesting authentication token")
+    async def _get_access_token(self):
+        _LOGGER.debug("Requesting authentication token (using username and password)")
         authentication = {"email": self._username, "password": self._password, "version": VERSION}
-        req = requests.post(AUTH_SERVER,
-                            headers={"x-lwrf-appid": "ios-01"},
-                            json=authentication)
-        _LOGGER.debug("Received response: {}".format(req))
-        if req.status_code == 200:
-            token = req.json()["tokens"]["access_token"]
-        else:
-            _LOGGER.warning("No authentication token (status_code '{}').".format(req.status_code))
-            raise ConnectionError("No authentication token: {}".format(req.text))
-        return token
+        async with self._session.post(AUTH_SERVER, headers={"x-lwrf-appid": "ios-01"}, json=authentication) as req:
+            _LOGGER.debug("Received response: {}".format(await req.json()))
+            if req.status == 200:
+                self._authtoken = (await req.json())["tokens"]["access_token"]
+            else:
+                _LOGGER.warning("No authentication token (status_code '{}').".format(req.status))
+                raise ConnectionError("No authentication token: {}".format(await req.text))
 
-class LWLink2Public:
+    #########################################################
+    # Convenience methods for non-async calls
+    #########################################################
 
-    def __init__(self, api_token, refresh_token):
-        self._api_token = api_token
-        self._refresh_token = refresh_token
-        self._access_token = None
-        self._token_expiry = None
-        self.featuresets = []
+    def _sendmessage(self, message):
+        return asyncio.get_event_loop().run_until_complete(self._async_sendmessage(message))
 
     def get_hierarchy(self):
         return asyncio.get_event_loop().run_until_complete(self.async_get_hierarchy())
 
+    def update_featureset_states(self):
+        return asyncio.get_event_loop().run_until_complete(self.async_update_featureset_states())
+
+    def write_feature(self, feature_id, value):
+        return asyncio.get_event_loop().run_until_complete(self.async_write_feature(feature_id, value))
+
+    def read_feature(self, feature_id):
+        return asyncio.get_event_loop().run_until_complete(self.async_read_feature(feature_id))
+
+    def turn_on_by_featureset_id(self, featureset_id):
+        return asyncio.get_event_loop().run_until_complete(self.async_turn_on_by_featureset_id(featureset_id))
+
+    def turn_off_by_featureset_id(self, featureset_id):
+        return asyncio.get_event_loop().run_until_complete(self.async_turn_off_by_featureset_id(featureset_id))
+
+    def set_brightness_by_featureset_id(self, featureset_id, level):
+        return asyncio.get_event_loop().run_until_complete(self.async_set_brightness_by_featureset_id(featureset_id, level))
+
+    def set_temperature_by_featureset_id(self, featureset_id, level):
+        return asyncio.get_event_loop().run_until_complete(self.async_set_temperature_by_featureset_id(featureset_id, level))
+
+    def connect(self):
+        return asyncio.get_event_loop().run_until_complete(self.async_connect())
+
+class LWLink2Public(LWLink2):
+
+    def __init__(self, username=None, password=None, auth_method="username", api_token=None, refresh_token=None):
+
+        self.featuresets = {}
+        self._authtoken = None
+
+        self._username = username
+        self._password = password
+        self._auth_method = auth_method
+        self._api_token = api_token
+        self._refresh_token = refresh_token
+
+        self._session = aiohttp.ClientSession()
+        self._token_expiry = None
+
+    #TODO add retries/error checking
+    #TODO make this async!
+    async def _async_getrequest(self, endpoint, _retry=1):
+        _LOGGER.debug("Sending API GET request to {}".format(endpoint))
+        req = requests.get(PUBLIC_API + endpoint,
+                           headers={"authorization": "bearer " + self._authtoken})
+        _LOGGER.debug("Received API response {}".format(req.json()))
+        return req.json()
+
+    async def _async_postrequest(self, endpoint, body="", _retry=1):
+        _LOGGER.debug("Sending API POST request to {}: {}".format(endpoint, body))
+        req = requests.post(PUBLIC_API + endpoint,
+                            headers={"authorization": "bearer " + self._authtoken}, json=body)
+        _LOGGER.debug("Received API response {}".format(req.json()))
+        return req.json()
+
     async def async_get_hierarchy(self):
 
-        self.featuresets = []
-        req = requests.get("https://publicapi.lightwaverf.com/v1/structures",
-                            headers={"authorization": "bearer " + self._access_token})
-        for struct in req.json()["structures"]:
-            req = requests.get("https://publicapi.lightwaverf.com/v1/structure/" + struct,
-                               headers={"authorization": "bearer " + self._access_token})
-            response = req.json()
+        self.featuresets = {}
+        req = await self._async_getrequest("structures")
+        for struct in req["structures"]:
+            response = await self._async_getrequest("structure/" + struct)
 
             for x in response["devices"]:
                 for y in x["featureSets"]:
@@ -409,7 +437,7 @@ class LWLink2Public:
                     new_featureset.name = x["name"]
 
                     for z in y["features"]:
-                        _LOGGER.debug("Adding device features {}".format(x))
+                        _LOGGER.debug("Adding device features {}".format(z))
                         new_featureset.features[z["type"]] = [z["featureId"], None]
                         if z["type"] == "switch":
                             new_featureset._switchable = True
@@ -419,7 +447,7 @@ class LWLink2Public:
                             new_featureset._climate = True
                         if z["type"] == "identify":
                             new_featureset._gen2 = True
-                    self.featuresets.append(new_featureset)
+                    self.featuresets[y["featureSetId"]] = new_featureset
 
         await self.async_update_featureset_states()
 
@@ -427,111 +455,35 @@ class LWLink2Public:
     async def async_register_callback(self, callback):
         pass
 
-    def update_featureset_states(self):
-        return asyncio.get_event_loop().run_until_complete(self.async_update_featureset_states())
-
     async def async_update_featureset_states(self):
-        for x in self.featuresets:
-            for y in x.features:
-                value = await self.async_read_feature(x.features[y][0])
-                x.features[y][1] = value
+        feature_list = []
 
-    def write_feature(self, feature_id, value):
-        return asyncio.get_event_loop().run_until_complete(self.async_write_feature(feature_id, value))
+        for dummy, x in self.featuresets.items():
+            for y in x.features:
+                feature_list.append({"featureId":x.features[y][0]})
+        body = {"features":feature_list}
+        req = await self._async_postrequest("features/read", body)
+
+        for featuresetid in self.featuresets:
+            for featurename in self.featuresets[featuresetid].features:
+                if self.featuresets[featuresetid].features[featurename][0] in req:
+                    self.featuresets[featuresetid].features[featurename][1] = req[self.featuresets[featuresetid].features[featurename][0]]
 
     async def async_write_feature(self, feature_id, value):
-        #TODO Check for expiry of auth token
         payload = {"value": value}
-        req = requests.post("https://publicapi.lightwaverf.com/v1/feature/" + feature_id,
-                               headers={"authorization": "bearer " + self._access_token},
-                            json=payload)
-        _LOGGER.debug("Received response: {}".format(req))
-
-    def read_feature(self, feature_id):
-        return asyncio.get_event_loop().run_until_complete(self.async_read_feature(feature_id))
+        await self._async_postrequest("feature/" + feature_id, payload)
 
     async def async_read_feature(self, feature_id):
-        # TODO Check for expiry of auth token
-        req = requests.get("https://publicapi.lightwaverf.com/v1/feature/" + feature_id,
-                               headers={"authorization": "bearer " + self._access_token})
-        return req.json()["value"]
-
-    def get_featureset_by_id(self, featureset_id):
-        for x in self.featuresets:
-            if x.featureset_id == featureset_id:
-                return x
-        return None
-
-    def get_featureset_by_featureid(self, feat_id):
-        for x in self.featuresets:
-            for y in x.features.values():
-                if y[0] == feat_id:
-                    return x
-        return None
-
-    def turn_on_by_featureset_id(self, featureset_id):
-        return asyncio.get_event_loop().run_until_complete(self.async_turn_on_by_featureset_id(featureset_id))
-
-    async def async_turn_on_by_featureset_id(self, featureset_id):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["switch"][0]
-        await self.async_write_feature(feature_id, 1)
-
-    def turn_off_by_featureset_id(self, featureset_id):
-        return asyncio.get_event_loop().run_until_complete(self.async_turn_off_by_featureset_id(featureset_id))
-
-    async def async_turn_off_by_featureset_id(self, featureset_id):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["switch"][0]
-        await self.async_write_feature(feature_id, 0)
-
-    def set_brightness_by_featureset_id(self, featureset_id, level):
-        return asyncio.get_event_loop().run_until_complete(self.async_set_brightness_by_featureset_id(featureset_id, level))
-
-    async def async_set_brightness_by_featureset_id(self, featureset_id, level):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["dimLevel"][0]
-        await self.async_write_feature(feature_id, level)
-
-    def set_temperature_by_featureset_id(self, featureset_id, level):
-        return asyncio.get_event_loop().run_until_complete(self.async_set_temperature_by_featureset_id(featureset_id, level))
-
-    async def async_set_temperature_by_featureset_id(self, featureset_id, level):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["targetTemperature"][0]
-        await self.async_write_feature(feature_id, int(level * 10))
-
-    def get_switches(self):
-        temp = []
-        for x in self.featuresets:
-            if x.is_switch():
-                temp.append((x.featureset_id, x.name))
-        return temp
-
-    def get_lights(self):
-        temp = []
-        for x in self.featuresets:
-            if x.is_light():
-                temp.append((x.featureset_id, x.name))
-        return temp
-
-    def get_climates(self):
-        temp = []
-        for x in self.featureset:
-            if x.is_climate():
-                temp.append((x.featureset_id, x.name))
-        return temp
+        req = await self._async_getrequest("feature/" + feature_id)
+        return req["value"]
 
     #########################################################
     # Connection
     #########################################################
 
-    def connect(self):
-        return asyncio.get_event_loop().run_until_complete(self.async_connect())
-
     async def async_connect(self, tries=0):
         try:
-            return await self._get_access_token()
+            self._get_access_token()
         except Exception as exp:
             retry_delay = min(2 ** (tries + 1), 120)
             _LOGGER.warning("Cannot connect (exception '{}'). Waiting {} seconds".format(exp, retry_delay))
@@ -539,17 +491,40 @@ class LWLink2Public:
             return await self.async_connect(tries + 1)
     #TODO distinguish failure on no token and don't retry
 
-    async def _get_access_token(self):
-        _LOGGER.debug("Requesting authentication token")
+    def _get_access_token(self):
+        if self._auth_method == "username":
+            self._get_access_token_username()
+        elif self._auth_method == "api":
+            self._get_access_token_api()
+        else:
+            raise ValueError("auth_method must be 'username' or 'api'")
+
+    def _get_access_token_username(self):
+        _LOGGER.debug("Requesting authentication token (using username and password)")
+        authentication = {"email": self._username, "password": self._password, "version": VERSION}
+        req = requests.post(AUTH_SERVER,
+                            headers={"x-lwrf-appid": "ios-01"},
+                            json=authentication)
+        _LOGGER.debug("Received response: {}".format(req.json()))
+        if req.status_code == 200:
+            token = req.json()["tokens"]["access_token"]
+            self._authtoken = token
+        else:
+            _LOGGER.warning("No authentication token (status_code '{}').".format(req.status_code))
+            raise ConnectionError("No authentication token: {}".format(req.text))
+
+    #TODO check for token expiry
+    def _get_access_token_api(self):
+        _LOGGER.debug("Requesting authentication token (using API key and refresh token)")
         authentication = {"grant_type": "refresh_token", "refresh_token": self._refresh_token}
         req = requests.post(PUBLIC_AUTH_SERVER,
                             headers={"authorization": "basic " + self._api_token},
                             json=authentication)
         _LOGGER.debug("Received response: {}".format(req))
         if req.status_code == 200:
-            self._access_token = req.json()["access_token"]
+            self._authtoken = req.json()["access_token"]
             self._refresh_token = req.json()["refresh_token"]
             self._token_expiry = datetime.datetime.now() + datetime.timedelta(seconds=req.json()["expires_in"])
         else:
             _LOGGER.warning("No authentication token (status_code '{}').".format(req.status_code))
-            raise ConnectionError("No authentication token: {}".format(req.text))
\ No newline at end of file
+            raise ConnectionError("No authentication token: {}".format(req.text))
